"""
Recip -- Sage package for using Shimura's reciprocity law

See the file main.sage for usage instructions.

#*****************************************************************************
# Copyright (C) 2010 -- 2012 Marco Streng <marco.streng@gmail.com>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#*****************************************************************************

To use this package, start Sage with the .sage files from this package in your
working directory. Then type::

    sage: load("main.sage")

This file contains functions for listing quartic CM-fields.

"""

def iterate_DAB(stopD, stopA, startD=5, startA=1, prime_D=False):
    """
    Given a range of positive discriminants D and integers A, returns all
    minimal triples DAB corresponding to primitive quartic CM-fields with D and
    A in the range.
    
    The number of such triples is proportional to
    (maxA^2 - minA^2) * (sqrt(maxD) - sqrt(minD)).
    
    If prime_D is true, restrict to D of the form p for a prime p=1 mod 4
    and 4*p for p a prime p!=1 mod 4.
    
    EXAMPLES::
    
        sage: list(iterate_DAB(10, 10))
        [[5, 5, 5], [5, 7, 11], [5, 9, 19], [8, 4, 2], [8, 6, 7], [8, 8, 14]]
        sage: list(iterate_DAB(20, 20, prime_D=True))
        [[5, 5, 5], [5, 7, 11], [5, 9, 19], [5, 10, 20], [5, 11, 29], [5, 12, 31], [5, 13, 41], [5, 14, 44], [5, 15, 55], [5, 15, 45], [5, 16, 59], [5, 17, 71], [5, 17, 61], [5, 18, 76], [5, 19, 89], [5, 19, 79], [8, 4, 2], [8, 6, 7], [8, 8, 14], [8, 10, 23], [8, 10, 17], [8, 12, 34], [8, 12, 18], [8, 14, 47], [8, 14, 41], [8, 14, 31], [8, 16, 62], [8, 16, 46], [8, 18, 79], [8, 18, 73], [8, 18, 63], [12, 6, 6], [12, 8, 13], [12, 10, 22], [12, 10, 13], [12, 12, 33], [12, 14, 46], [12, 14, 37], [12, 14, 22], [12, 16, 61], [12, 16, 37], [12, 18, 78], [12, 18, 69], [12, 18, 33], [13, 5, 3], [13, 9, 17], [13, 10, 12], [13, 12, 23], [13, 13, 39], [13, 13, 13], [13, 15, 53], [13, 16, 51], [13, 17, 69], [13, 17, 43], [13, 18, 68], [13, 18, 29], [13, 19, 87], [13, 19, 61], [17, 5, 2], [17, 11, 26], [17, 12, 19], [17, 13, 38], [17, 15, 52], [17, 15, 18], [17, 16, 47], [17, 17, 68], [17, 17, 34], [17, 19, 86]]
        sage: l = list(iterate_DAB(30, 30))
        sage: l = [(DAB, CM_Field(DAB)) for DAB in l]
        sage: l = [(DAB, K, K.ideal(2).factor()) for (DAB, K) in l]
        sage: l = [(DAB, K) for (DAB, K, f) in l if len(f) == 1 and f[0][1] == 1]
        sage: l = [(DAB, K.class_number()/K.real_field().class_number()) for (DAB, K) in l]
        sage: [(DAB, h) for (DAB, h) in l if h > 3]
        [([21, 25, 109], 4), ([29, 25, 149], 5)]
    """
    # d is D//4
    for D in range(startD, stopD):
        if prime_D:
            if D % 4 == 1 and not ZZ(D).is_prime():
                continue
            elif D % 4 == 0 and not ZZ(D/4).is_prime():
                continue
        if not is_fundamental_discriminant(D):
            continue
        K0 = QuadraticField(D, 'a')
        for DAB in iterate_DAB_given_D(D, stopA, startA, K0):
            yield DAB


def iterate_DAB_given_D(D, stopA, startA=1, K0=None):
    """
    Given a real quadratic discriminant and a range of integers A,
    returns all minimal triples DAB corresponding to primitive quartic
    CM-fields with this D and in the A-range.
    
    The number of such triples is proportional to
    1/2 (maxA^2 - minA^2) / sqrt(D).
    
    K0, if given, should be the quadratic field with discriminant D.
    """
    # D is known, also m^2 D = A^2 - 4B,
    # so if D is even, then so is A
    if D % 2 == 0:
        e = 2
        if startA % 2 == 1:
            startA = startA + 1
    else:
        e = 1
    
    for A in range(startA, stopA, e):
        for DAB in iterate_DAB_given_DA(D, A, K0):
            yield DAB


def iterate_DAB_given_DA(D, A, K0):
    """
    Given a real quadratic discriminant and a positive integer A,
    returns all minimal triples DAB corresponding to primitive quartic
    CM-fields with this D and in the A-range.

    The number of such triples is proportional to A/sqrt(D).

    K0, if given, should be the quadratic field with discriminant D.
    """
    if K0 is None:
        K0 = QuadraticField(D,'a')
    # if D is even, then both D and A^2 are divisible by 4
    # if D is odd, then it is 1 mod 4, so m and A have the same parity
    # Take steps for m of length e, and start with start.
    if D % 2 == 0:
        e = 1
    else:
        e = 2
    start = A % 2
    if start == 0:
        start = e
    # Now m^2 D = A^2 - 4B < A^2, so m < A/sqrt(D), so m <= floor(A/sqrt(D))
    for m in range(start, A/sqrt(D)+1, e):
        DAB = DAm_to_DAB(D, A, m)
        if not DAB[2].is_square(): # equivalent to primitive CM-type
            min_dab = DAB_to_minimal(DAB, K0, m)
            if min_dab == DAB:
                yield DAB


def DAm_to_DAB(D, A, m):
    """
    Given D a fundamental discriminant, A a positive integer, and m
    a positive integer with m^2 D < A^2 and m^2 * D == A^2 mod 4,
    returns [D,A,B] with m^2*D = A^2 - 4B
    """
    return [D, A, ZZ((A**2-m**2*D)/4)]


def DAB_to_minimal(DAB, K0=None, m=None, check=False):
    """
    Given a triple DAB, finds a triple representing the same field, but such
    that A is minimal (and B is minimal with that A).
    
    K0, if given, should be the quadratic field with discriminant D.

    m, if given, must be a positive integer satisfying m^2 * D = A^2 - 4B
    """
    [D,A,B] = DAB
    if K0 is None:
        K0 = QuadraticField(D,'a')
    if m is None:
        m = ZZ(sqrt((A**2-4*B)/D))
    if check:
        assert m**2*D == A**2-4*B
    sqrtD = K0.gen()
    
    # We compute an ideal b such that beta is in b if and only if
    # beta*alpha is integral, where alpha = sqrt(-A+m*sqrtD)
    # Note first that beta*alpha is integral if and only if
    # beta^2*alpha^2 is integral. So let's compute alpha^2:
    alphasq = (-A+m*sqrtD)/2
    a = K0.ideal(alphasq)
    f = a.factor()
    b = K0.ideal(1)
    for (p,e) in f:
        b = b*p**-((e/2).floor())
    # Now beta is in b if and only if beta^2*alpha^2 is integral.
    bas = b.basis()
    
    # Next, we create a matrix such that the square of the Euclidean norm of
    # beta*alpha for beta = x*bas[0]+y*bas[2] is (x,y) M (x,y)^t
    #M = Matrix([[(beta1*beta2*alphasq).trace() for beta1 in bas] \
    #             for beta2 in bas])
    Q = QuadraticForm(ZZ, 2, [-(bas[0]**2*alphasq).trace(),\
                              -2*(bas[0]*bas[1]*alphasq).trace(),\
                              -(bas[1]**2*alphasq).trace()])
    R, T = Q.reduced_binary_form1()
    (a,b,c) = R.coefficients()
    assert a == Q(T.column(0))
    assert c == Q(T.column(1))
    s = [T.column(0)]
    if a == c:
        s.append(T.column(1))
        if b == a:
            s.append(T.column(0)-T.column(1))
        elif b == -a:
            s.append(T.column(0)+T.column(1))
    
    # Now s contains, up to sign, all elements beta with alpha' = beta*alpha of minimal
    # Euclidean norm. The square of this Euclidean norm is a = -trace(alpha'^2) = A'
    
    DABS = []
    for beta in s:
        b = -ZZ((((beta[0]*bas[0]+beta[1]*bas[1])**2*alphasq * 2 + a)**2 - a**2) / 4)
        DABS.append([D,a,b])
    DABS.sort()
    return DABS[0]

